/*********************************************************************
*               SEGGER MICROCONTROLLER GmbH & Co KG                  *
*       Solutions for real time microcontroller applications         *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2015  SEGGER Microcontroller GmbH & Co KG         *
*                                                                    *
*       www.segger.com     Support: support@segger.com               *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.06b                                            *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_SAMD20_CMSIS.c for Atmel SAMD20 and Atmel SAMD20-XPRO

Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#include "samd20.h"     // Device specific header file, contains CMSIS

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock frequency settings
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS SystemCoreClock
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #define OS_PCLK_TIMER OS_FSYS     /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART OS_FSYS      /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000u)
#endif

/*********************************************************************
*
*       Configuration of communication to OSView
*/
#ifndef   OS_VIEW_ENABLE            // Global enable of communication
  #define OS_VIEW_ENABLE    (1)     // Default: on
#endif

#ifndef   OS_VIEW_USE_UART          // If set, UART will be used for communication
  #define OS_VIEW_USE_UART (1)      // Default: 0 => memory access is used
#endif                              // if OS_VIEW_ENABLE is on

/****** End of configurable options *********************************/

/*********************************************************************
*
*       Check configuration
*
**********************************************************************
*/
#ifndef   DEBUG     /* Should normally be defined as project option */
  #define DEBUG  (0)
#endif


/*********************************************************************
*
*       UART settings for OSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to OSView, please define it to be -1
*       Currently the standard code enables SERCOM3 per default
*/
#ifndef   OS_UART
  #define OS_UART (3)
#endif

#ifndef   OS_BAUDRATE
  #define OS_BAUDRATE (115200u)
#endif

/*********************************************************************
*
*       JLINKMEM and UART settings for OSView
*
*       Automatically generated from configuration settings
*/
#define OS_USE_JLINKMEM   (OS_VIEW_ENABLE && (OS_VIEW_USE_UART == 0))

#define OS_UART_USED      (OS_VIEW_ENABLE && (OS_VIEW_USE_UART != 0) && (OS_UART == 3))

#if OS_USE_JLINKMEM
  #include "JLINKMEM.h"
#endif

/****** MAP UART initialization function ****************************/

#if  (OS_UART_USED || OS_USE_JLINKMEM)
  #define OS_COM_INIT() OS_COM_Init()
#else
  #define OS_COM_INIT()
#endif

/*********************************************************************
*
*       Vector table
*/
#ifdef __ICCARM__
  #define __Vectors    __vector_table
#else
  extern unsigned char __Vectors;
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
#define NVIC_VTOR         (*(volatile OS_U32*) (0xE000ED08uL))
#define NVIC_HFSR         (*(volatile OS_U32*) (0xE000ED2CuL))

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

#if OS_USE_JLINKMEM
  // Size of the communication buffer for JLINKMEM
const OS_U32 OS_JLINKMEM_BufferSize = 32u;
#else
const OS_U32 OS_JLINKMEM_BufferSize = 0;     // Communication not used
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
* Function description
*   Returns the current hardware timer count value
*
* Return value
*   Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
	return SysTick->VAL;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
* Function description
*   Returns if the hardware timer interrupt pending flag is set
*
* Return value
*   == 0; Interrupt pending flag not set
*   != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
	return SCB->ICSR & SCB_ICSR_PENDSTSET_Msk;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       SysTick_Handler()
*
* Function description
*   This is the code that gets called when the processor receives a
*   _SysTick exception. SysTick is used as OS timer tick.
*/
#ifdef __cplusplus
extern "C" {
#endif
void SysTick_Handler(void);      // Avoid warning, Systick_Handler is not prototyped in any CMSIS header
#ifdef __cplusplus
}
#endif
void SysTick_Handler(void) {
  OS_EnterNestableInterrupt();
  OS_TICK_Handle();
  #if OS_USE_JLINKMEM
    JLINKMEM_Process();
  #endif
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       OS_InitHW()
*
*       Initialize the hardware (timer) required for the OS to run.
*       May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  //
  // Structure with information about timer frequency, tick frequency, etc.
  // for micro second precise system time.
  // SysTimerConfig.TimerFreq will be set later, thus it is initialized with zero.
  //
  OS_SYSTIMER_CONFIG SysTimerConfig = {0, OS_TICK_FREQ, 0, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

  OS_IncDI();
  //
  // We assume, the PLL and core clock was already set by the SystemInit() function
  // which was called from the startup code
  // Therefore, we don't have to initialize any hardware here,
  // we just ensure that the system clock variable is updated and then
  // set the periodic system timer tick for embOS.
  //
  SystemCoreClockUpdate();                             // Update the system clock variable (might not have been set before)
  if (SysTick_Config (OS_PCLK_TIMER / OS_TICK_FREQ)) { // Setup SysTick Timer for 1 msec interrupts
    while (1);                                         // Handle Error
  }
  //
  // Initialize NVIC vector base address. Might be necessary for RAM targets or application not running from 0
  //
  NVIC_VTOR = (OS_U32)&__Vectors;
  //
  // Set the interrupt priority for the system timer to 2nd lowest level to ensure the timer can preempt PendSV handler
  //
  NVIC_SetPriority(SysTick_IRQn, (1u << __NVIC_PRIO_BITS) - 2u);
  //
  // Setup values for usec precise system time functions
  //
  SysTimerConfig.TimerFreq = SystemCoreClock;
  OS_Config_SysTimer(&SysTimerConfig);

  OS_COM_INIT();
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*       Please note:
*       This is basically the "core" of the idle loop.
*       This core loop can be changed, but:
*       The idle loop does not have a stack of its own, therefore no
*       functionality should be implemented that relies on the stack
*       to be preserved. However, a simple program loop can be programmed
*       (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {     // Idle loop: No task is ready to execute
  while (1) {
    #if ((OS_USE_JLINKMEM == 0) && (DEBUG == 0))     // Enter CPU halt mode when not in DEBUG build and J-Link communication not used
      #ifdef __ICCARM__  // IAR
        __asm(" wfi");
      #endif
      #ifdef __CC_ARM    // KEIL
        __wfi();
      #endif
      #ifdef __GNUC__    // GCC
        __asm(" wfi");
      #endif
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*       This routine is required for task-info via OSView or high
*       resolution time measurement functions.
*       It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U32 Cnt;

  Time = OS_GetTime32();
  Cnt  = (OS_PCLK_TIMER/1000u) - SysTick->VAL;
  //
  // Check if timer interrupt pending ...
  //
  if (SCB->ICSR & SCB_ICSR_PENDSTSET_Msk) {
    Cnt = (OS_PCLK_TIMER/1000u) - SysTick->VAL;     // Interrupt pending, re-read timer and adjust result
    Time++;
  }
  return ((OS_PCLK_TIMER/1000u) * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*       Convert Cycles into micro seconds.
*
*       If your clock frequency is not a multiple of 1 MHz,
*       you may have to modify this routine in order to get proper
*       diagnostics.
*
*       This routine is required for profiling or high resolution time
*       measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return Cycles/(OS_PCLK_TIMER/1000000u);
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

#if OS_USE_JLINKMEM                    // Communication via JTAG / SWD

static void _JLINKMEM_OnRx(OS_U8 Data);
static void _JLINKMEM_OnTx(void);
static OS_INT _JLINKMEM_GetNextChar(void);

/*********************************************************************
*
*       _JLINKMEM_OnRx()
*/
static void _JLINKMEM_OnRx(OS_U8 Data) {
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _JLINKMEM_OnTx()
*/
static void _JLINKMEM_OnTx(void) {
  OS_OnTx();
}

/*********************************************************************
*
*       _JLINKMEM_GetNextChar()
*/
static OS_INT _JLINKMEM_GetNextChar(void) {
  return OS_COM_GetNextChar();
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize memory access for OSView
*/
void OS_COM_Init(void) {
  JLINKMEM_SetpfOnRx(_JLINKMEM_OnRx);
  JLINKMEM_SetpfOnTx(_JLINKMEM_OnTx);
  JLINKMEM_SetpfGetNextChar(_JLINKMEM_GetNextChar);
}

/*********************************************************************
*
*       OS_COM_Send1()
*       Send one character via memory
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKMEM_SendChar(c);
}

#elif OS_UART_USED   // Communication via UART, can not be implemented generic
//
// S3_UART_TX: Pin PA24
// S3_UART_RX: PIN_PA25
//
#define OS_BAUD_VALUE          (65536uLL - ((65536uLL * OS_BAUDRATE * 16uLL) / OS_PCLK_UART))

#define SERCOM3_BASE_ADDR      (0x42001400u)
#define OS_UART_RX_PIN         (10u)
#define OS_UART_TX_PIN         (11u)
#define SERCOM3_IRQn           (10u)

#define SERCOM_CTRLA           (*(volatile OS_U32*)(SERCOM3_BASE_ADDR + 0x00))  
#define SERCOM_CTRLB           (*(volatile OS_U32*)(SERCOM3_BASE_ADDR + 0x04))
#define SERCOM_BAUD            (*(volatile OS_U16*)(SERCOM3_BASE_ADDR + 0x0A))
#define SERCOM_INTENCLR        (*(volatile OS_U8*) (SERCOM3_BASE_ADDR + 0x0C))
#define SERCOM_INTENSET        (*(volatile OS_U8*) (SERCOM3_BASE_ADDR + 0x0D))
#define SERCOM_INTFLAG         (*(volatile OS_U8*) (SERCOM3_BASE_ADDR + 0x0E))
#define SERCOM_STATUS          (*(volatile OS_U16*)(SERCOM3_BASE_ADDR + 0x10))
#define SERCOM_DATA            (*(volatile OS_U16*)(SERCOM3_BASE_ADDR + 0x18))

#define SERCOM_LSB_FIRST       (1u << 30u)                                     // LSB first
#define SERCOM_TXC             (1u << 1u)                                      // Transmit Complete Interrupt Enable
#define SERCOM_RXC             (1u << 2u)                                      // Receive Complete Interrupt Enable
#define SERCOM_RXPO            (3u << 20u)                                     // SERCOM PAD[3] is used for data reception
#define SERCOM_TXPO            (1u << 16u)                                     // SERCOM PAD[2] is used for data transmit
#define SERCOM_ENABLE          (1u << 1u)                                      // Enable SERCOM bit
#define SERCOM_TXEN            (1u << 16u)                                     // Transmitter Enable
#define SERCOM_RXEN            (1u << 17u)                                     // Receiver Enable
#define SERCOM_MODE_INTCLK     (1u << 2u)                                      // Internal clock mode

#define SERCOM_ERROR_FLAGS     (0x07)                                          // BUFOVF, FERR and PERR
#define SERCOM_INTFLAG_RXC     (1u << 2u)                                      // RXC Receive complete
#define SERCOM_INTFLAG_TXC     (1u << 1u)                                      // TXC Transmit complete
#define SERCOM_STATUS_SYNCBUSY (1u << 15)                                      // Synchronization Busy

#define APBCMASK               (*(volatile OS_U32*)(0x40000420))
#define APBCMASK_SERCOM3       (1u << 5u)
#define APBCMASK_PAC2          (1u << 0u)

#define GCLK_BASE_ADDR         (0x40000C00u)
#define GCLK_CLKCTRL           (*(volatile OS_U16*)(GCLK_BASE_ADDR + 0x02))

#define PORT_BASE_ADDR         (0x41004400u)
#define PORT_PINCFGA24         (*(volatile OS_U8*)(PORT_BASE_ADDR + 0x58)) 
#define PORT_PINCFGA25         (*(volatile OS_U8*)(PORT_BASE_ADDR + 0x59)) 
#define PORT_PMUX12            (*(volatile OS_U8*)(PORT_BASE_ADDR + 0x3C))

#define PORT_PMUXEN            (1u << 0u)
#define PORT_INEN              (1u << 1u)
#define PORT_PULLEN            (1u << 2u)
#define PERIPERAL_FUNC_C       (0x02u)

/*********************************************************************
*
*       SERCOM3_Handler()
*       Interrupt handler for SERCOM3 interrupts
*/
void SERCOM3_Handler(void) {
  volatile OS_U32 Dummy __attribute__ ((unused)); // suppress unused-but-set warning, it is required for clearing register
  unsigned int UsartStatus;
  
  OS_EnterNestableInterrupt();

  UsartStatus = SERCOM_STATUS;                    // Examine status register

  if (SERCOM_INTFLAG & SERCOM_INTFLAG_RXC) {      // Data received?
    if (UsartStatus & SERCOM_ERROR_FLAGS) {       // Any error?
      Dummy = SERCOM_DATA;                        // => Discard data
      SERCOM_STATUS = SERCOM_ERROR_FLAGS; 
    } else {
      OS_OnRx(SERCOM_DATA);                       // Process actual byte
    }
  }
  if (SERCOM_INTFLAG & SERCOM_INTFLAG_TXC) {      // Check Tx status => Send next character
    if (OS_OnTx()) {                              // No more characters to send?
      SERCOM_INTENCLR = SERCOM_TXC;               // Disable further tx interrupts
    }
  }

  OS_LeaveNestableInterrupt();
}
 
/*********************************************************************
*
*       OS_COM_Send1()
*       Send one character via memory
*/
void OS_COM_Send1(OS_U8 c) {
  SERCOM_DATA = c;
  SERCOM_INTENSET = SERCOM_TXC;  // Enable tx interrupt
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize memory access for OSView
*/
void OS_COM_Init(void) {
  //
  // Enable peripheral
  APBCMASK |= APBCMASK_SERCOM3;
  //
  //
  // Enable clock   
  //
  GCLK_CLKCTRL |= (1u << 14u) | 0x10u;  // SERCOM3, use generic clock generator 0
  GCLK_CLKCTRL |= (1u << 14u) | 0x0Cu;  // SERCOM_slow, use generic clock generator 0
  //
  // Set port pins
  //
  PORT_PINCFGA24   = PORT_PMUXEN;
  PORT_PINCFGA25   = PORT_PMUXEN;
  PORT_PMUX12      = PERIPERAL_FUNC_C | (PERIPERAL_FUNC_C << 4u);
  //
  // Set uart settings
  //
  SERCOM_CTRLA     = SERCOM_LSB_FIRST | SERCOM_MODE_INTCLK | SERCOM_RXPO | SERCOM_TXPO;
  SERCOM_CTRLB     = SERCOM_TXEN | SERCOM_RXEN;
  SERCOM_BAUD      = OS_BAUD_VALUE;
  SERCOM_INTENSET  = SERCOM_RXC;
  SERCOM_CTRLA    |= SERCOM_ENABLE;
  //
  // Install USART Handler with preemption level one above lowest level to ensure communication during PendSV
  //
  NVIC_SetPriority(SERCOM3_IRQn, (1u << __NVIC_PRIO_BITS) - 2u);
  NVIC_EnableIRQ(SERCOM3_IRQn);
}

#else  /*  No communication or UART selected, using dummies */

void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           /* Avoid compiler warning */
  OS_COM_ClearTxActive();  /* Let the OS know that Tx is not busy */
}

#endif /*  OS_UART_USED  */

/****** Final check of configuration ********************************/

#ifndef OS_UART_USED
  #error "OS_UART_USED has to be defined"
#endif

#if (OS_TICK_FREQ != 1000u)
  #error "OS_TICK_FREQ != 1000, ensure OS_GetTime_Cycles() and OS_ConvertCycles2us() work."
#endif

/*********************************************************************
*
*       HardFault_Handler()
*
*       Using a debugger and setting a breakpoint here allows to analyze
*       the cause of the hardfault.
*       Normally, the CPU should not arrive here.
*
*   NOTE:
*       Do not modify the code. The function might fail if local
*       variables are used.
*/
static volatile unsigned int _Continue;

void HardFault_Handler(void) {
	_Continue = 0u;
	//
	// When stuck here, change the variable value to != 0 in order to step out
	//
	while (_Continue == 0u);
}

/****** End Of File *************************************************/
